/* Error & Logging functions

 Copyright  2010 Travis Robinson. <libusbdotnet@gmail.com>
 website: http://sourceforge.net/projects/libusb-win32
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU Lesser General Public License as published by 
 the Free Software Foundation; either version 2 of the License, or 
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
 along with this program; if not, please visit www.gnu.org.
*/
 
#include "error.h"
#include <errno.h>
#include <string.h>
#include <stdio.h>

#if IS_DRIVER
	#ifdef __GNUC__
		#define OBJ_KERNEL_HANDLE       0x00000200L
		#include <ddk/usb100.h>
		#include <ddk/usbdi.h>
		#include <ddk/winddk.h>
		#include "usbdlib_gcc.h"
	#else
		#include <ntddk.h>
	#endif
#else
	#include <windows.h>
#endif

#define USB_ERROR_BEGIN			500000

#ifndef LOG_APPNAME
#define LOG_APPNAME "LOG_APPNAME define missing"
#endif

#define GetLogLevel(UsbLogLevel) ((UsbLogLevel & LOG_LEVEL_MASK)>LOG_LEVEL_MAX?LOG_LEVEL_MAX:UsbLogLevel & LOG_LEVEL_MASK)
#define GetLogOuput(LogOutputType) (LogOutputType>0?(_LOG_OUTPUT_TYPE & LogOutputType):1)

void usb_err_v	(const char* function, const char* format, va_list args);
void usb_wrn_v	(const char* function, const char* format, va_list args);
void usb_msg_v	(const char* function, const char* format, va_list args);
void usb_dbg_v	(const char* function, const char* format, va_list args);

void usb_log_v	(enum USB_LOG_LEVEL level, const char* function, const char* format, va_list args);
void _usb_log	(enum USB_LOG_LEVEL level, const char* app_name, const char* function, const char* format, ...);
void _usb_log_v	(enum USB_LOG_LEVEL level, const char* app_name, const char* function, const char* format, va_list args);

void WriteDriverLogEntry(char* message, const int message_length);

static void usb_log_def_handler(enum USB_LOG_LEVEL level, 
								const char* app_name, 
								const char* prefix, 
								const char* func, 
								const int app_prefix_func_end,
								char* message,
								const int message_length);

#define STRIP_PREFIX(stringSrc, stringPrefix) \
	(strstr(stringSrc,stringPrefix)==stringSrc?stringSrc+strlen(stringPrefix):stringSrc)

static const char *log_level_string[LOG_LEVEL_MAX+1] =
{
    "off",
    "error",
    "warning",
    "info",
    "debug",

    "unknown",
};

static const char *skipped_function_prefix_list[] =
{
    "usb_registry_",
    "usb_",
	NULL
};

char usb_error_str[LOGBUF_SIZE] = "";
int usb_error_errno = 0;

#if (IS_DEBUG_MODE)
int __usb_log_level = LOG_DEBUG;
#else
int __usb_log_level = LOG_OFF;
#endif

usb_error_type_t usb_error_type = USB_ERROR_TYPE_NONE;

const char** skipped_function_prefix = skipped_function_prefix_list;

#if !IS_DRIVER

char *usb_strerror(void)
{
    switch (usb_error_type)
    {
    case USB_ERROR_TYPE_NONE:
        return "No error";
    case USB_ERROR_TYPE_STRING:
        return usb_error_str;
    case USB_ERROR_TYPE_ERRNO:
        if (usb_error_errno > -USB_ERROR_BEGIN)
            return strerror(usb_error_errno);
        else
            /* Any error we don't know falls under here */
            return "Unknown error";
    }

    return "Unknown error";
}

/* returns Windows' last error in a human readable form */
const char *usb_win_error_to_string(void)
{
    static char tmp[LOGBUF_SIZE];

    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
                  LANG_USER_DEFAULT, tmp, sizeof(tmp) - 1, NULL);

    return tmp;
}


int usb_win_error_to_errno(void)
{
    switch (GetLastError())
    {
    case ERROR_SUCCESS:
        return 0;
    case ERROR_INVALID_PARAMETER:
        return EINVAL;
    case ERROR_SEM_TIMEOUT:
    case ERROR_OPERATION_ABORTED:
        return ETIMEDOUT;
    case ERROR_NOT_ENOUGH_MEMORY:
        return ENOMEM;
    default:
        return EIO;
    }
}

#endif

void usb_err(const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    usb_err_v(function, format, args);
    va_end(args);
}
void usb_wrn(const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    usb_wrn_v(function, format, args);
    va_end(args);
}

void usb_msg(const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    usb_msg_v(function, format, args);
    va_end(args);
}

void usb_dbg(const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    usb_dbg_v(function, format, args);
    va_end(args);
}

void usb_log(enum USB_LOG_LEVEL level, const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    usb_log_v(level, function, format, args);
    va_end(args);
}

void usb_err_v(const char* function, const char* format, va_list args)
{
    usb_log_v(LOG_ERROR, function, format, args);
}

void usb_wrn_v(const char* function, const char* format, va_list args)
{
    usb_log_v(LOG_WARNING, function, format, args);
}

void usb_msg_v(const char* function, const char* format, va_list args)
{
    usb_log_v(LOG_INFO, function, format, args);
}

void usb_dbg_v(const char* function, const char* format, va_list args)
{
    usb_log_v(LOG_DEBUG, function, format, args);
}

void usb_log_v(enum USB_LOG_LEVEL level, const char* function, const char* format, va_list args)
{
    _usb_log_v(level, LOG_APPNAME, function, format, args);
}

void _usb_log(enum USB_LOG_LEVEL level, const char* app_name, const char* function, const char* format, ...)
{
    va_list args;
    va_start(args, format);
    _usb_log_v(level, app_name, function, format, args);
    va_end(args);
}

void _usb_log_v(enum USB_LOG_LEVEL level,
                const char* app_name,
                const char* function,
                const char* format,
                va_list args)
{

    char local_buffer[LOGBUF_SIZE];
    int totalCount, count;
    const char* prefix;
    const char* func;
    char* buffer;
    int masked_level;
	const char** skip_list;
	int app_prefix_func_end;

	masked_level = GetLogLevel(level);

    if (__usb_log_level < masked_level && masked_level != LOG_ERROR) return;
    buffer = local_buffer;
    totalCount = 0;
    count = 0;

    if (masked_level > LOG_LEVEL_MAX) masked_level = LOG_LEVEL_MAX;

    if ((level & LOG_RAW) == LOG_RAW)
    {
        count = _vsnprintf(buffer, LOGBUF_SIZE-1, format, args);
        if (count > 0)
        {
            buffer += count;
            totalCount += count;
        }
    }
    else
    {
        prefix = log_level_string[masked_level];
		func = function;

		if (func)
		{
			// strip some prefixes to shorten function names
			skip_list=skipped_function_prefix;
			while(*skip_list && ((func)) && func[0])
			{
				func = STRIP_PREFIX(func,skip_list[0]);
				skip_list++;
			}
		}

		if(!func) func="none";

        // print app name, level string and short function name
        count = _snprintf(buffer, (LOGBUF_SIZE-1), "%s:%s [%s] ", app_name, prefix, func);
        if (count > 0)
        {
			app_prefix_func_end = count;
            buffer += count;
            totalCount += count;
            count = _vsnprintf(buffer, (LOGBUF_SIZE-1) - totalCount, format, args);
            if (count > 0)
            {
                buffer += count;
                totalCount += count;
            }
        }
    }
    if (count < 0)
        totalCount = LOGBUF_SIZE - 1;

    // make sure its null terminated
    local_buffer[totalCount] = 0;

#if (!IS_DRIVER)
    if (masked_level == LOG_ERROR)
    {
        // if this is an error message then store it
        strncpy(usb_error_str, local_buffer, totalCount);
        usb_error_str[totalCount] = '\0';
        usb_error_type = USB_ERROR_TYPE_STRING;
    }
#endif

    if (__usb_log_level >= masked_level)
    {
		usb_log_def_handler(level, app_name, prefix, func, app_prefix_func_end, local_buffer, totalCount);
    }
}

void usb_log_set_level(enum USB_LOG_LEVEL level)
{
    __usb_log_level = level > LOG_LEVEL_MAX ? LOG_LEVEL_MAX : level;
}

int usb_log_get_level()
{
    return __usb_log_level;
}

/* Default log handler
*/
static void usb_log_def_handler(enum USB_LOG_LEVEL level, 
								const char* app_name, 
								const char* prefix, 
								const char* func, 
								const int app_prefix_func_end,
								char* message,
								const int message_length)
{
#if GetLogOuput(LOG_OUTPUT_TYPE_FILE) && !IS_DRIVER
	FILE* file;
#endif

#if GetLogOuput(LOG_OUTPUT_TYPE_DBGPRINT)
	DbgPrint("%s",message);
#endif

#if GetLogOuput(LOG_OUTPUT_TYPE_STDERR)
	fprintf(stderr, message);
#endif

#if GetLogOuput(LOG_OUTPUT_TYPE_DEBUGWINDOW)
	OutputDebugStringA(message);
#endif

// TODO: Kernel driver must use ZwCreateFile
#if GetLogOuput(LOG_OUTPUT_TYPE_FILE)
	#if IS_DRIVER
		WriteDriverLogEntry(message,message_length);
	#else
		file = fopen(LOG_FILE_PATH,"a");
		if (file)
		{
			fwrite(message,1,strlen(message),file);
			fflush(file);
			fclose(file);
		}
	#endif
#endif

#if GetLogOuput(LOG_OUTPUT_TYPE_MSGBOX)
	if (GetLogLevel(level)==LOG_ERROR)
	{
		message[app_prefix_func_end-1]='\0';
		MessageBoxA(NULL,message+strlen(message),message,MB_OK|MB_ICONERROR);
	}
#endif
}

#if IS_DRIVER
void WriteDriverLogEntry(char* message, const int message_length)
{
	HANDLE                      logFileHandle=NULL;       // DataFile handle.
	OBJECT_ATTRIBUTES           objectAttributes;			// Used for opening file.
	UNICODE_STRING              uniFileName;
	ANSI_STRING					ansiFilename;
	IO_STATUS_BLOCK             ioStatusBlock;
	NTSTATUS                    ntStatus = STATUS_SUCCESS;

	RtlInitAnsiString(&ansiFilename,LOG_FILE_PATH);
	ntStatus = RtlAnsiStringToUnicodeString(&uniFileName,&ansiFilename,TRUE);
	if (!NT_SUCCESS(ntStatus))
		return;

	// Create data file.
	InitializeObjectAttributes(&objectAttributes,
		&uniFileName,
		OBJ_CASE_INSENSITIVE|OBJ_KERNEL_HANDLE,
		NULL,
		NULL);

	ntStatus = ZwCreateFile(&logFileHandle,
							SYNCHRONIZE | FILE_APPEND_DATA,
							&objectAttributes,
							&ioStatusBlock,
							NULL,
							FILE_ATTRIBUTE_NORMAL,
							0,
							FILE_OPEN_IF,
							FILE_NON_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT,
							NULL,
							0);

	if (NT_SUCCESS(ntStatus))
	{

		ntStatus = ZwWriteFile( logFileHandle,
								NULL,
								NULL,
								NULL,
								&ioStatusBlock,
								message,
								message_length,
								NULL,
								NULL);

		if (NT_SUCCESS(ntStatus))
		{
		}

		ZwClose(logFileHandle);
		logFileHandle = NULL;
	}

	RtlFreeUnicodeString(&uniFileName);
}
#endif
